"""
Configurations that are required by the Docker Build, these include the main configurations which
contains details that are common across multiple builds and the build configurations which contains
details for one particular build cycle.
"""