# the path to the build context on the container. This determines where the specified build context
# folder on the build machine will be copied on the container.
BUILD_CONTEXT_DST_PATH = "/tmp/build-context"

# the default path for the configuration file
CONFIG_FILE_PATH = "~/.docker/build-config.yml"
