"""
Groups together the functionality that is required to communicate with the Docker daemon that the
build tool will use to perform the build. The package contains the commands that can be sent to the
daemon through the rest APIs and also other functionality required for such communication.
"""